/* approx.c 
 * Approximation algorithms for RFLCS
 */

#include "approx.h"

/* Given two sequences x and y and their respective sizes m and n,
 * stores in z a LCS of x and y and returns its size.
 * (Introduction to Algorithms, Cormen et. al, p. 350-355)
 */
int lcs_cormen (int *x, int *y, int *z, int m, int n)
{
  int i, j, k;
  int **c;
  char **b;

  c = (int **) malloc_chk((m + 1) * sizeof (int *));
  for (i = 0; i < m + 1; i++)
    c[i] = (int *) malloc_chk((n + 1) * sizeof (int));

  b = (char **) malloc_chk((m + 1) * sizeof (char *));
  for (i = 0; i < m + 1; i++)
    b[i] = (char *) malloc_chk((n + 1) * sizeof (char));

  for (i = 1; i <= m; i++)
    c[i][0] = 0;
  for (j = 0; j <= n; j++)
    c[0][j] = 0;

  for (i = 1; i <= m; i++) {
    for (j = 1; j <= n; j++) {
      if (x[i-1] == y[j-1]) {
	c[i][j] = c[i-1][j-1] + 1;
	b[i][j] = 'd';
      }
      else if (c[i-1][j] >= c[i][j-1]) {
	c[i][j] = c[i-1][j];
	b[i][j] = 'u';
      }
      else {
	c[i][j] = c[i][j-1];
	b[i][j] = 'l';
      }
    }
  }

  k = c[m][n] - 1;
  i = m;
  j = n;

  while (i != 0 && j != 0) {
    if (b[i][j] == 'd') {
      z[k] = x[i-1];
      i--;
      j--;
      k--;
    }
    else if (b[i][j] == 'u')
      i--;
    else
      j--;
  }

  k = c[m][n];

  for (i = 0; i < m + 1; i++) {
    free(b[i]);
    free(c[i]);
  }
  free(b);
  free(c);

  return k;
}

/* Given x and y, computes a LCS of x and y and throws away all repetitions
 * in the obtained LCS. Stores the resulting sequence in z and returns its
 * size.
 */
int rflcs_approx1 (int *x, int *y, int *z, int m, int n, int asize)
{
  int i, j, k;
  int *zaux, *repeat;

  zaux = (int *) malloc_chk(MIN(m, n) * sizeof (int));
  repeat = (int *) malloc_chk(asize * sizeof (int));

  k = lcs_cormen(x, y, zaux, m, n);

  for (i = 0; i < asize; i++)
    repeat[i] = 0;

  j = 0;
  for (i = 0; i < k; i++) {
    if (repeat[(unsigned int) zaux[i]] == 0) {
      z[j] = zaux[i];
      j++;
      repeat[(unsigned int) zaux[i]] = 1;
    }
  }

  free(zaux);
  free(repeat);

  return j;
}

/* If type is 1:
 * Given x and y, for each symbol a, picks uniformly at random one occurrence
 * of a in x or y, whichever has less occurrences of a, and deletes all
 * others, resulting in x' and y'. Computes a LCS using x' and y'. Stores the 
 * resulting sequence in z and returns its size.
 *
 * If type is 2:
 * Given x and y, picks at random one number in the interval [0, 1] and
 * uses it to decide which occurrence of each symbol will remain.
 * Computes a LCS using the resulting sequences, stores it in z, and
 * returns its size.
 *
 * Repeats PROB_TIMES times.
 */
int rflcs_approx2 (int *x, int *y, int *z, int m, int n, int type, int asize)
{
  int i, j, k, kmax, maux, naux;
  int *xcount, *ycount;
  int *xchoice, *ychoice;
  int *xaux, *yaux, *zaux;
  double **prob;
  
  xcount = (int *) malloc_chk(asize * sizeof (int));
  ycount = (int *) malloc_chk(asize * sizeof (int));
  xchoice = (int *) malloc_chk(asize * sizeof (int));
  ychoice = (int *) malloc_chk(asize * sizeof (int));
  xaux = (int *) malloc_chk(m * sizeof (int));
  yaux = (int *) malloc_chk(n * sizeof (int));
  zaux = (int *) malloc_chk(MIN(m, n) * sizeof (int));

  kmax = -1;

  count_seq(xcount, ycount, x, y, m, n, asize);

  if (type == 3) {
    prob = (double **) malloc_chk(asize * sizeof (double *));
    for (i = 0; i < asize; i++)
      prob[i] = (double *) malloc_chk(MAX(m, n) * sizeof (double));
    prepare_3(x, y, m, n, xcount, ycount, asize, prob);
  }

  for (i = 0; i < PROB_TIMES; i++) {

    for (j = 0; j < asize; j++)
      xchoice[j] = ychoice[j] = 0;

    if (type == 1)
      choice_1(xcount, ycount, xchoice, ychoice, asize);
    else if (type == 2)
      choice_2(xcount, ycount, xchoice, ychoice, asize);
    else
      choice_3(xcount, ycount, xchoice, ychoice, asize, prob);

    maux = choose_seq(x, xaux, xchoice, m);
    naux = choose_seq(y, yaux, ychoice, n);

    k = lcs_cormen(xaux, yaux, zaux, maux, naux);

    if (k > kmax) {
      kmax = k;
      for (j = 0; j < k; j++)
        z[j] = zaux[j];
    }
  }

  free(xcount);
  free(ycount);
  free(xchoice);
  free(ychoice);
  free(xaux);
  free(yaux);
  free(zaux);

  if (type == 3) {
    for (i = 0; i < asize; i++)
      free(prob[i]);
    free(prob);
  }

  return kmax;
}

/* Derandomized version of rflcs_approx2 with type = 2.
 */
int rflcs_approx2_derand (int *x, int *y, int *z, int m, int n, int asize)
{
  int j, k, kmax, maux, naux;
  int *xcount, *ycount, *drcount;
  int *xchoice, *ychoice;
  int *xaux, *yaux, *zaux;
  unsigned int lcmx, lcmy, lcm, range;
  
  xcount = (int *) malloc_chk(asize * sizeof (int));
  ycount = (int *) malloc_chk(asize * sizeof (int));
  drcount = (int *) malloc_chk(asize * sizeof (int));
  xchoice = (int *) malloc_chk(asize * sizeof (int));
  ychoice = (int *) malloc_chk(asize * sizeof (int));
  xaux = (int *) malloc_chk(m * sizeof (int));
  yaux = (int *) malloc_chk(n * sizeof (int));
  zaux = (int *) malloc_chk(MIN(m, n) * sizeof (int));

  kmax = 0;

  count_seq(xcount, ycount, x, y, m, n, asize);

  lcmx = lcm_array(xcount, asize);
  lcmy = lcm_array(ycount, asize);
  lcm = lcm_2(lcmx, lcmy);

  range = 0;
  while (range < lcm) {
    range = choice_2_derand(xcount, ycount, xchoice, ychoice, asize, 
			    range, lcm);

    maux = choose_seq(x, xaux, xchoice, m);
    naux = choose_seq(y, yaux, ychoice, n);

    k = lcs_cormen(xaux, yaux, zaux, maux, naux);

    if (k > kmax) {
      kmax = k;
      for (j = 0; j < k; j++)
	z[j] = zaux[j];
    }
  }

  free(xcount);
  free(ycount);
  free(xchoice);
  free(ychoice);
  free(xaux);
  free(yaux);
  free(zaux);

  return kmax;
}


/* Randomly chooses the amount of each symbol, generating a random number
 * for each one.
 */
void choice_1 (int *xcount, int *ycount, int *xchoice, int *ychoice, int asize)
{
  int i;

  for (i = 0; i < asize; i++) {
    if (xcount[i] != 0)
      xchoice[i] = 1 + (int) (xcount[i] * (rand() / (RAND_MAX + 1.0)));
    else if (ycount[i] != 0)
      ychoice[i] = 1 + (int) (ycount[i] * (rand() / (RAND_MAX + 1.0)));
  }
}

/* Randomly chooses the amount of each symbol, generating only one random
 * number.
 */
void choice_2 (int *xcount, int *ycount, int *xchoice, int *ychoice, int asize)
{
  int i;
  double t;

  t = ((double) rand()) / RAND_MAX;

  for (i = 0; i < asize; i++) {
    if (xcount[i] != 0)
      xchoice[i] = 1 + (int) (xcount[i] * t);
    else if (ycount[i] != 0)
      ychoice[i] = 1 + (int) (ycount[i] * t);
  }
}

/* Chooses the amount of each symbol based on k.
 */
unsigned int choice_2_derand (int *xcount, int *ycount, int *xchoice, 
			      int *ychoice, int asize, unsigned int k, 
			      unsigned int lcm)
{
  int i;
  unsigned int tmp, jump, jumpmin;

  for (i = 0; i < asize; i++) {
    xchoice[i] = (xcount[i] > 0) ? 1 : 0;
    ychoice[i] = (ycount[i] > 0) ? 1 : 0;
  }

  if (k == 0)
    return 1;

  jumpmin = lcm;
  for (i = 0; i < asize; i++) {
    if (xcount[i] > 1) {
      tmp = 1 + k / (lcm / xcount[i]);
      xchoice[i] = tmp;
      jump = tmp * (lcm / xcount[i]);
      if (jump < jumpmin)
	jumpmin = jump;
    }
    if (ycount[i] > 1) {
      tmp = 1 + k / (lcm / ycount[i]);
      ychoice[i] = tmp;
      jump = tmp * (lcm / ycount[i]);
      if (jump < jumpmin)
	jumpmin = jump;
    }
  }

  return jumpmin;
}

/* Constructs an LP problem and runs simplex, then builds a matrix of
 * probabilities for each pair of symbols based on it.
 */
void prepare_3 (int *x, int *y, int m, int n, int *xcount, int *ycount,
                int asize, double **prob)
{
  glp_prob *lp;
  glp_smcp *lpparm;
  crossdata_t *cd;

  cd = malloc_crossdata(m, n, asize);

  construct_model_ip();
  construct_data_ip(x, y, m, n, asize, cd);

  lp = lpx_read_model(MODPATH, DATPATH, NULL);

  lpparm = (glp_smcp *) malloc_chk(sizeof (glp_smcp));
  glp_init_smcp(lpparm);
  glp_simplex(lp, lpparm);

  construct_prob_lp(lp, x, y, m, n, xcount, ycount, asize, cd, prob);

  glp_delete_prob(lp);

  free_crossdata(cd, m);
}

/* Builds a matrix of probabilities for each pair of symbols based on simplex.
 */
void construct_prob_lp (glp_prob *lp, int *x, int *y, int m, int n, 
			int *xcount, int *ycount, int asize, crossdata_t *cd, 
			double **prob)
{
  int i, j, k;
  double sum;

  for (i = 0; i < asize; i++)
    for (j = 0; j < MAX(m, n); j++)
      prob[i][j] = 0.0;

  j = 1;
  for (i = 0; i < asize; i++) {
    k = 0;
    while (j <= cd->csize && x[cd->crossx[j-1]] == i) {
      prob[i][k] += glp_get_col_prim(lp, j);
      j++;
      if (xcount[i] != 0) {
        if (cd->crossx[j-1] > cd->crossx[j-2])
          k++;
      }
      else if (ycount[i] != 0) {
        k++;
        if (cd->crossy[j-1] < cd->crossy[j-2])
          k = 0;
      }
    }
  }

  for (i = 0; i < asize; i++) {
    if (xcount[i] != 0) {
      sum = 0.0;
      for (j = 0; j < xcount[i]; j++)
	sum += prob[i][j];
      for (j = 0; j < xcount[i]; j++)
	prob[i][j] /= sum;
    }
    else if (ycount[i] != 0) {
      sum = 0.0;
      for (j = 0; j < ycount[i]; j++)
	sum += prob[i][j];
      for (j = 0; j < ycount[i]; j++)
	prob[i][j] /= sum;
    }
  }
}

/* Chooses the amount of each symbol based on the matrix of probabilities.
 */
void choice_3 (int *xcount, int *ycount, int *xchoice, int *ychoice, 
               int asize, double **prob)
{
  int i, k;
  double t;

  for (i = 0; i < asize; i++) {
    if (xcount[i] != 0) {
      t = ((double) rand()) / RAND_MAX;
      for (k = 0; k < xcount[i]; k++) {
        if (t < prob[i][k])
          break;
        else t -= prob[i][k];
      }
      xchoice[i] = k + 1;
    }
    else if (ycount[i] != 0) {
      t = ((double) rand()) / RAND_MAX;
      for (k = 0; k < ycount[i]; k++) {
        if (t < prob[i][k])
          break;
        else t -= prob[i][k];
      }
      ychoice[i] = k + 1;
    }
  }
}

/* Counts each symbol in x and y and stores it in xcount and ycount.
 * If ycount[i] > xcount[i], xcount[i] = 0, and vice-versa.
 */
void count_seq (int *xcount, int *ycount, int *x, int *y, int m, int n, 
		int asize)
{
  int i;

  for (i = 0; i < asize; i++) {
    xcount[i] = 0;
    ycount[i] = 0;
  }

  for (i = 0; i < m; i++)
    xcount[(unsigned int) x[i]]++;
  for (i = 0; i < n; i++)
    ycount[(unsigned int) y[i]]++;

  for (i = 0; i < asize; i++) {
    if (xcount[i] > ycount[i])
      xcount[i] = 0;
    else ycount[i] = 0;
  }
}

/* Given a sequence x and its length m, creates a sequence xaux based on
 * choice. If choice[a] = k, then the k-th symbol a is put on xaux.
 * Returns the length of xaux.
 */
int choose_seq (int *x, int *xaux, int *choice, int m)
{
  int i, j;

  j = 0;
  for (i = 0; i < m; i++) {
    if (choice[(unsigned int) x[i]] == 0) {
      xaux[j] = x[i];
      j++;
    }
    else if (choice[(unsigned int) x[i]] == 1) {
      xaux[j] = x[i];
      j++;
      choice[(unsigned int) x[i]] = -1;
    }
    else
      choice[(unsigned int) x[i]]--;
  }

  return j;
}
