/* ip.c
 * Integer programming functions
 */

#include "ip.h"

int rflcs_ip (int *xo, int *yo, int *z, int mo, int no, int asize, 
	      graph_t *graph)
{
  int k, m, n, asizeel;
  int *x, *y;
  glp_prob *lp;
  crossdata_t *cd;

  x = (int *) malloc_chk(mo * sizeof (int));
  y = (int *) malloc_chk(no * sizeof (int));

  printf("Preprocessing sequences...\n");
  asizeel = preprocess_rflcs(xo, yo, x, y, mo, no, &m, &n, asize);
  printf("x = ");
  print_sequence(x, m);
  printf("y = ");
  print_sequence(y, n);
  printf("Length of sequence x: %d\n", m);
  printf("Length of sequence y: %d\n", n);
  printf("Preprocessed alphabet size: %d\n", asizeel);
  printf("\n");

  cd = malloc_crossdata(m, n, asize);

  if (graph != NULL)
    construct_model_max_ip(graph);
  else construct_model_ip();
  construct_data_ip(x, y, m, n, asize, cd);
  
  lp = lpx_read_model(MODPATH, DATPATH, NULL);

  lpx_set_int_parm(lp, LPX_K_ROUND, 1); /* Rounds zeros */
  if (graph == NULL)
    lpx_set_int_parm(lp, LPX_K_USECUTS, LPX_C_ALL);  /* Cutting planes */

  lpx_set_real_parm(lp, LPX_K_TMLIM, TIME_LIMIT);

  lpx_intopt(lp);

  k = convert_lp_to_sequence(lp, x, y, z, cd, 1);

  glp_delete_prob(lp);

  free_crossdata(cd, m);

  return k;
}

/* Constructs and writes an IP data file in GNU MathProg format.
 */
void construct_data_ip (int *x, int *y, int m, int n, int asize, 
			crossdata_t *cd)
{
  FILE *datafile;
  int i, j, k;

  datafile = fopen(DATPATH, "w");
  if (datafile == NULL) {
    fprintf(stderr, "Error: Unable to write IP data file.\n");
    exit(EXIT_FAILURE);
  }

  fprintf(datafile, "set S :=");
  for (i = 0; i < asize; i++)
    fprintf(datafile, " %d", i);
  fprintf(datafile, ";\n");

  cd->csize = 0;
  for (k = 0; k < asize; k++) {
    fprintf(datafile, "set E[%d] :=", k);
    cd->scross[k] = cd->csize;
    for (i = 0; i < m; i++)
      if (x[i] == k)
        for (j = 0; j < n; j++)
          if (y[j] == k) {
            fprintf(datafile, " (%d,%d)", i, j);
            cd->crossx[cd->csize] = i;
            cd->crossy[cd->csize] = j;
	    cd->revcross[i][j] = cd->csize;
            (cd->csize)++;
          }
    fprintf(datafile, ";\n");
  }

  fprintf(datafile, "\nend;\n");

  fclose(datafile);
}

/* Constructs and writes an IP model file in GNU MathProg format using 
 * crossings.
 */
void construct_model_ip ()
{
  FILE *modfile;

  modfile = fopen(MODPATH, "w");
  if (modfile == NULL) {
    fprintf(stderr, "Error: Unable to write IP model file.\n");
    exit(EXIT_FAILURE);
  }

  fprintf(modfile, "/* set of symbols */\n");
  fprintf(modfile, "set S;\n\n");

  fprintf(modfile, "set E{a in S} dimen 2;\n\n");

  fprintf(modfile, "set EE := setof{a in S, (i,j) in E[a]} (i,j);\n\n");

  fprintf(modfile, "/* (i,j) and (k,l) cross if i < k and j > l */\n");
  fprintf(modfile, "set crossing := setof{a in S, b in S, (i,j,k,l) in "
          "E[a] cross E[b]: i < k and j > l and a <> b} (i,j,k,l);\n\n");

  fprintf(modfile, "/* z[i,j] = 1 iff x[i] and y[i] are aligned in a "
          "RFLCS of x and y */\n");
  fprintf(modfile, "var z{(i,j) in EE}, binary;\n\n");

  fprintf(modfile, "s.t. repfree{a in S}: sum{(i,j) in E[a]} "
          "z[i,j] <= 1;\n\n");

  fprintf(modfile, "s.t. nocross{(i,j,k,l) in crossing}: "
          "z[i,j] + z[k,l] <= 1;\n\n");

  fprintf(modfile, "maximize rflcs: sum{(i,j) in EE} z[i,j];\n\n");

  fprintf(modfile, "solve;\n\n");
  
  fprintf(modfile, "end;\n");

  fclose(modfile);
}

/* Constructs and writes an IP model file in GNU MathProg format using
 * maximal crossings based on graph.
 */
void construct_model_max_ip (graph_t *graph)
{
  int *stack, top, k;
  FILE *modfile;

  modfile = fopen(MODPATH, "w");
  if (modfile == NULL) {
    fprintf(stderr, "Error: Unable to write IP model file.\n");
    exit(EXIT_FAILURE);
  }

  stack = (int *) malloc_chk((graph->nv + 2) * sizeof (int));

  top = 0;
  k = 0;

  fprintf(modfile, "/* set of symbols */\n");
  fprintf(modfile, "set S;\n\n");

  fprintf(modfile, "set E{a in S} dimen 2;\n\n");

  fprintf(modfile, "set EE := setof{a in S, (i,j) in E[a]} (i,j);\n\n");

  fprintf(modfile, "/* z[i,j] = 1 iff x[i] and y[i] are aligned in a "
          "RFLCS of x and y */\n");
  fprintf(modfile, "var z{(i,j) in EE}, binary;\n\n");

  fprintf(modfile, "s.t. repfree{a in S}: sum{(i,j) in E[a]} "
          "z[i,j] <= 1;\n\n");

  search_graph(modfile, graph, stack, &top, graph->nv + 1, &k);
  fprintf(modfile, "\n");

  fprintf(modfile, "maximize rflcs: sum{(i,j) in EE} z[i,j];\n\n");

  fprintf(modfile, "solve;\n\n");
  
  fprintf(modfile, "end;\n");

  fclose(modfile);
}

/* Constructs and writes an IP model file in GNU MathProg format only
 * with repetition constraints.
 */
void construct_model_reponly_ip ()
{
  FILE *modfile;

  modfile = fopen(MODPATH, "w");
  if (modfile == NULL) {
    fprintf(stderr, "Error: Unable to write IP model file.\n");
    exit(EXIT_FAILURE);
  }

  fprintf(modfile, "/* set of symbols */\n");
  fprintf(modfile, "set S;\n\n");

  fprintf(modfile, "set E{a in S} dimen 2;\n\n");

  fprintf(modfile, "set EE := setof{a in S, (i,j) in E[a]} (i,j);\n\n");

  fprintf(modfile, "/* z[i,j] = 1 iff x[i] and y[i] are aligned in a "
          "RFLCS of x and y */\n");
  fprintf(modfile, "var z{(i,j) in EE}, binary;\n\n");

  fprintf(modfile, "s.t. repfree{a in S}: sum{(i,j) in E[a]} "
	  "z[i,j] <= 1;\n\n");

  fprintf(modfile, "maximize rflcs: sum{(i,j) in EE} z[i,j];\n\n");

  fprintf(modfile, "solve;\n\n");
  
  fprintf(modfile, "end;\n");

  fclose(modfile);
}

void search_graph (FILE *modfile, graph_t *graph, int *stack, int *top, 
                   int v, int *k)
{
  int i;

  if (v == graph->nv) {
    printf("building nocross%d, top = %d\n", *k, *top);
    fprintf(modfile, "s.t. nocross%d: ", (*k)++);
    for (i = 0; i < *top - 1; i++)
      fprintf(modfile, "+ z[%d,%d] ", graph->vmap[stack[i]][0], 
             graph->vmap[stack[i]][1]);
    fprintf(modfile, "<= 1;\n");
  }
  else {
    for (i = 0; i < graph->nv + 1; i++) {
      if (graph->adj[v][i] == -1) {
        stack[(*top)++] = i;
        search_graph(modfile, graph, stack, top, i, k);
        (*top)--;
      }
    }
  }
}

/* Given a LP problem with a valid solution, creates a sequence z with
 * the associated sequence. mip is a flag that is 1 iff the solution
 * obtained is from a MIP problem.
 */
int convert_lp_to_sequence (glp_prob *lp, int *x, int *y, int *z,
			    crossdata_t *cd, int mip)
{
  int i, j, k;

  k = 0;
  for (i = 1; i <= cd->csize; i++) {
    if (mip == 1)
      j = (int) (glp_mip_col_val(lp, i) + 0.5);
    else
      j = (int) (glp_get_col_prim(lp, i) + 0.5);
    if (j == 1) {
      if (x[cd->crossx[i-1]] != y[cd->crossy[i-1]]) {
	fprintf(stderr, "Error: Unable to convert LP solution to valid "
		"RFLCS solution.\n");
	exit(EXIT_FAILURE);
      }
      z[k++] = cd->crossx[i-1];
    }
  }

  qsort(z, k, sizeof (int), int_compare);
  for (i = 0; i < k; i++)
    z[i] = x[z[i]];

  return k;
}

/* Given a valid subsequence s of x and y, converts it to an array of 
 * variables of an associated LP problem.
 */
void convert_sequence_to_vars (int *x, int *y, int *s, int k, double *vars, 
			       crossdata_t *cd)
{
  int i, xi, yi;

  for (i = 1; i <= cd->csize; i++)
    vars[i] = 0;
  
  xi = yi = 0;
  for (i = 0; i < k; i++) {
    while (x[xi] != s[i])
      xi++;
    while (y[yi] != s[i])
      yi++;
    vars[cd->revcross[xi][yi] + 1] = 1;
  }
}
