/* metapost.c
 * Metapost printing
 */

#include "metapost.h"

/* Prints a metapost file representing the x and y sequences and an alignment
 * that defines z.
 */
void print_metapost (int *x, int *y, int *z, int m, int n, int k, 
                     FILE *mpfile, int version, int color, int asize)
{
  int i, xi, yi;

  fprintf(mpfile, "beginfig(%d);\n\n", version);

  fprintf(mpfile, "pickup pencircle scaled 4pt;\n\n");

  if (color == 0) {
    fprintf(mpfile, "for i = 0 upto %d:\n", m);
    fprintf(mpfile, "  draw (10 + 10 * i, 10);\n");
    fprintf(mpfile, "endfor;\n\n");
    
    fprintf(mpfile, "for i = 0 upto %d:\n", n);
    fprintf(mpfile, "  draw (10 + 10 * i, 60);\n");
    fprintf(mpfile, "endfor;\n\n");
  }
  else {
    for (i = 0; i < m; i++) {
      fprintf(mpfile, "draw (%d,%d)", 10 + 10 * i, 10);
      print_metapost_withcolor(mpfile, color, asize, x[i]);
      fprintf(mpfile, ";\n");
     }
    for (i = 0; i < n; i++) {
      fprintf(mpfile, "draw (%d,%d)", 10 + 10 * i, 60);
      print_metapost_withcolor(mpfile, color, asize, y[i]);
      fprintf(mpfile, ";\n");
    }
  }
  
  fprintf(mpfile, "pickup pencircle scaled 1pt;\n\n");

  xi = 0;
  yi = 0;
  for (i = 0; i < k; i++) {
    while (x[xi] != z[i])
      xi++;
    while (y[yi] != z[i])
      yi++;
    fprintf(mpfile, "draw (%d,10)--(%d,60)", 10 + xi * 10, 10 + yi * 10);
    print_metapost_withcolor(mpfile, color, asize, z[i]);
    fprintf(mpfile, ";\n");
  }
  fprintf(mpfile, "\n");

  /* Changes Metapost formatting for fonts to work on external display */
  fprintf(mpfile, "prologues := 2;\n");
  fprintf(mpfile, "label.rt(\"length = %d\", (10,0));\n", k);
  fprintf(mpfile, "\n");

  fprintf(mpfile, "endfig;\n\n");
}

/* Chooses a color and prints the withcolor metapost command.
 */
void print_metapost_withcolor (FILE *mpfile, int color, int asize, int c)
{
  double r, g, b;

  if (color != 0) {
    if (color == 1)
      r = g = b = 0.8 * c / asize;
    else
      find_hue(c, asize, &r, &g, &b);
    fprintf(mpfile, " withcolor (%f,%f,%f)", r, g, b);
  }
}

/* Prints a metapost file representing a graph of the sequences x and y
 * where its vertices are alignments and the edges indicate whether
 * the alignments cross or not, depending on its arguments.
 */
void print_metapost_graph (int *x, int *y, int m, int n, FILE *mpfile, 
			   int version, int color, int asize, int cross,
			   int clean)
{
  int i, j, k, l;
  int nlimit, jinit, tmp;

  fprintf(mpfile, "beginfig(%d);\n\n", version);

  fprintf(mpfile, "draw (0,0)--(%d,0);\n", 10 + 10 * m);
  fprintf(mpfile, "draw (0,0)--(0,%d);\n", 10 + 10 * n);
  fprintf(mpfile, "draw (0,%d)--(%d,%d);\n", 10 + 10 * n, 10 + 10 * m, 
	  10 + 10 * n);
  fprintf(mpfile, "draw (%d,0)--(%d,%d);\n", 10 + 10 * m, 10 + 10 * m, 
	  10 + 10 * n);

  fprintf(mpfile, "pickup pencircle scaled 4pt;\n\n");

  if (color != 0) {
    for (i = 0; i < m; i++) {
      fprintf(mpfile, "draw (%d,-10)", 10 + 10 * i);
      print_metapost_withcolor(mpfile, color, asize, x[i]);
      fprintf(mpfile, ";\n");
    }
    for (j = 0; j < n; j++) {
      fprintf(mpfile, "draw (-10,%d)", 10 + 10 * j);
      print_metapost_withcolor(mpfile, color, asize, y[j]);
      fprintf(mpfile, ";\n");
    }
  }
  fprintf(mpfile, "\n");

  for (i = 0; i < m; i++)
    for (j = 0; j < n; j++) {
      if (x[i] == y[j]) {
	nlimit = n;
	tmp = 0;
	if (cross == 0) {
	  for (k = i + 1; k < m; k++) {
	    for (l = j + 1; l < nlimit; l++)
	      if (x[k] == y[l]) {
		fprintf(mpfile, "drawarrow (%d,%d)--(%d,%d) withpen pencircle "
			"scaled 0.5pt;\n", 10 + 10 * i, 10 + 10 * j, 
			10 + 10 * k, 10 + 10 * l);
		if (clean == 1 && tmp == 0)
		  tmp = l + 1;
	      }
	    if (tmp != 0) {
	      nlimit = tmp;
	      tmp = 0;
	    }
	  }
	}
	else {
	  jinit = j + 1;
	  for (k = i; k >= 0; k--) {
	    for (l = jinit; l < nlimit; l++)
	      if (x[k] == y[l]) {
		fprintf(mpfile, "drawarrow (%d,%d)--(%d,%d) withpen pencircle "
			"scaled 0.5pt;\n", 10 + 10 * i, 10 + 10 * j, 
			10 + 10 * k, 10 + 10 * l);
		if (clean == 1) {
		  tmp = l;
		  break;
		}
	      }
	    if (tmp != 0) {
	      nlimit = tmp;
	      tmp = 0;
	    }
	    jinit = j;
	  }
	}
	fprintf(mpfile, "draw (%d,%d)", 10 + 10 * i, 10 + 10 * j);
	print_metapost_withcolor(mpfile, color, asize, x[i]);
	fprintf(mpfile, ";\n");
      }
    }
  fprintf(mpfile, "\n");

  fprintf(mpfile, "endfig;\n\n");

}

/* Stores in r, g, b a color value based on n.
 */
void find_hue (int n, int asize, double *r, double *g, double *b)
{
  double c, decpart;
  int intpart;

  c = 6 * n / ((double) asize);
  intpart = (int) c;
  decpart = c - intpart;
  switch (intpart) {
    case 0:
      *r = 1; *g = decpart; *b = 0;
      break;
    case 1:
      *r = 1 - decpart; *g = 1; *b = 0;
      break;
    case 2:
      *r = 0; *g = 1; *b = decpart;
      break;
    case 3:
      *r = 0; *g = 1 - decpart; *b = 1;
      break;
    case 4:
      *r = decpart; *g = 0; *b = 1;
      break;
    case 5:
      *r = 1; *g = 0; *b = 1 - decpart;
      break;
    default:
      fprintf(stderr, "Error: Unable to select colors.\n");
      exit(EXIT_FAILURE);    
  }
  *r *= 0.8;
  *g *= 0.8;
  *b *= 0.8;
}

void print_metapost_graph_done (int *x, int *y, int m, int n, FILE *mpfile,
				int version, int color, int asize, int cross,
				graph_t *graph)
{
  int i, j;

  if (cross < 0 || cross > 1) {
    fprintf(stderr, "Error: invalid parameter on printing metapost.\n");
    exit(EXIT_FAILURE);
  }
  if (cross == 0)
    cross = 1;
  else
    cross = -1;

  fprintf(mpfile, "beginfig(%d);\n\n", version);

  fprintf(mpfile, "draw (0,0)--(%d,0);\n", 10 + 10 * m);
  fprintf(mpfile, "draw (0,0)--(0,%d);\n", 10 + 10 * n);
  fprintf(mpfile, "draw (0,%d)--(%d,%d);\n", 10 + 10 * n, 10 + 10 * m, 
	  10 + 10 * n);
  fprintf(mpfile, "draw (%d,0)--(%d,%d);\n", 10 + 10 * m, 10 + 10 * m, 
	  10 + 10 * n);

  fprintf(mpfile, "pickup pencircle scaled 4pt;\n\n");

  if (color != 0) {
    for (i = 0; i < m; i++) {
      fprintf(mpfile, "draw (%d,-10)", 10 + 10 * i);
      print_metapost_withcolor(mpfile, color, asize, x[i]);
      fprintf(mpfile, ";\n");
    }
    for (j = 0; j < n; j++) {
      fprintf(mpfile, "draw (-10,%d)", 10 + 10 * j);
      print_metapost_withcolor(mpfile, color, asize, y[j]);
      fprintf(mpfile, ";\n");
    }
  }
  fprintf(mpfile, "\n");

  for (i = 0; i < graph->nv; i++)
    for (j = 0; j < graph->nv; j++) {
      if (graph->adj[i][j] == cross) {
	fprintf(mpfile, "drawarrow (%d,%d)--(%d,%d) withpen pencircle scaled "
		"0.5pt;\n", 10 + 10 * graph->vmap[i][0], 
		10 + 10 * graph->vmap[i][1], 10 + 10 * graph->vmap[j][0], 
		10 + 10 * graph->vmap[j][1]);
      }
    }

  for (i = 0; i < graph->nv; i++) {
    fprintf(mpfile, "draw (%d,%d)", 10 + 10 * graph->vmap[i][0], 
	    10 + 10 * graph->vmap[i][1]);
    print_metapost_withcolor(mpfile, color, asize, x[graph->vmap[i][0]]);
    fprintf(mpfile, ";\n");
  }
  fprintf(mpfile, "\n");

  fprintf(mpfile, "endfig;\n\n");
}
