/* seqgen.c
 * Sequence generation
 */

#include "seqgen.h"

/* Generates sequences x and y according to method.
 */
void generate_sequence (int *x, int *y, int *m, int *n, int asize, 
                        int genmethod)
{
  double *prob;

  switch (genmethod) {
    case 1:
      generate_sequence_uniform(x, asize, *m);
      generate_sequence_uniform(y, asize, *n);
      break;
    case 2:
      prob = (double *) malloc_chk(asize * sizeof (double));
      generate_prob_1(prob, asize);
      generate_sequence_prob(x, asize, *m, prob);
      generate_sequence_prob(y, asize, *n, prob);
      break;
    case 3:
      *m = generate_sequence_repetition(x, asize, *m);
      *n = generate_sequence_repetition(y, asize, *n);
      break;
    default:
      fprintf(stderr, "Error: Sequence generation method invalid.\n");
      exit(EXIT_FAILURE);
  }
}

/* Given an array x of length n and an integer asize, generates on x a 
 * uniform sequence composed of numbers from 0 to asize - 1.
 */
void generate_sequence_uniform (int *x, int asize, int n)
{
  int i;

  for (i = 0; i < n; i++)
    x[i] = (int) asize * (rand() / (RAND_MAX + 1.0));
}

/* Given an array x of length n and an array accprob of length asize,
 * generates on x a sequence based on the accumulated probabilities of
 * each element stored on accprob, which must be an ordered array
 * greater or equal than 0.0 and the last element should be 1.0.
 */
void generate_sequence_accprob (int *x, int asize, int n, double *accprob)
{
  int i;
  double r;

  for (i = 0; i < n; i++) {
    r = ((double) rand()) / RAND_MAX;
    x[i] = bsearchh(r, accprob, asize);
  }
}

/* Generates on x a sequence based on the probability distribution of its
 * elements stored on prob.
 */
void generate_sequence_prob (int *x, int asize, int n, double *prob)
{
  double *accprob;

  accprob = (double *) malloc_chk(asize * sizeof (double));

  create_accprob(prob, accprob, asize);
  generate_sequence_accprob(x, asize, n, accprob);
}

/* Creates an accumulated probability distribution from a probability
 * distribution.
 */
void create_accprob (double *prob, double *accprob, int asize)
{
  int i;

  accprob[0] = prob[0];
  for (i = 1; i < asize - 1; i++) {
    accprob[i] = accprob[i-1] + prob[i];
  }
  accprob[asize-1] = 1.0;
}

/* Generates a probability distribution such that the higher half of the
 * alphabet is five times the lower half.
 */
void generate_prob_1 (double *prob, int asize)
{
  int i, total;

  total = 0;
  for (i = 0; i < asize; i++)
    if (i < asize / 2) {
      prob[i] = 1.0;
      total += 1;
    }
    else {
      prob[i] = 5.0;
      total += 5;
    }

  for (i = 0; i < asize; i++)
    prob[i] /= total;
}

/* Generates a sequence on x where each symbol repeats at most rep times.
 */
int generate_sequence_repetition (int *x, int asize, int rep)
{
  int i, j, n, r;

  n = 0;
  for (i = 0; i < asize; i++) {
    r = 1 + (int) rep * (rand() / (RAND_MAX + 1.0));
    for (j = 0; j < r; j++) {
      x[n] = i;
      n++;
    }
  }

  shuffle_sequence(x, n);

  return n;
}

/* Randomly switches places of the elements of x.
 */
void shuffle_sequence (int *x, int n)
{
  int i, r, tmp;

  for (i = 0; i < n - 1; i++) {
    r = i + (int) (n - i) * (rand() / (RAND_MAX + 1.0));
    tmp = x[i];
    x[i] = x[r];
    x[r] = tmp;
  }
}

/* Loads sequences x and y from the file seqfname. The file must have
 * a number indicating alphabet size, the sizes of the sequences x and
 * y, and then the sequences x and y.
 */
void load_sequence (char seqfname[], int *x, int *y, int *m, int *n, 
                    int *asize)
{
  int i;
  FILE *seqfile;

  seqfile = fopen(seqfname, "r");
  if (seqfile == NULL) {
    fprintf(stderr, "Error: Unable to open sequences file.\n");
    exit(EXIT_FAILURE);
  }

  fscanf(seqfile, "%d %d %d", asize, m, n);
  
  for (i = 0; i < *m; i++)
    fscanf(seqfile, "%d", &(x[i]));

  for (i = 0; i < *n; i++)
    fscanf(seqfile, "%d", &(y[i]));

  fclose(seqfile);
}
