/* solver.c
 * An exact solver based on branch and cut
 */

#include "solver.h"

/* A hack to allow time limit. */
struct tms global_t_start;
struct tms global_t_end;

/*
 * Steps:
 * 1. Obtain an upper bound with LCS
 * 2. Preprocess sequences to eliminate unnecessary symbols
 * 3. Obtain lower bounds with approximation algorithms
 * 4. Run branch and cut algorithm
 */

int rflcs_bc (int *xo, int *yo, int *z, int mo, int no, int asize)
{
  int i, j, k, m, n;
  int upper, lower, asizeel;
  int *x, *y, *zaux;
  int *xcount, *ycount;
  info_t *info;
  glp_prob *lp;
  glp_smcp *lpparm;
  glp_iocp *ipparm;
  crossdata_t *cd;

  x = (int *) malloc_chk(mo * sizeof (int));
  y = (int *) malloc_chk(no * sizeof (int));
  zaux = (int *) malloc_chk(MIN(mo, no) * sizeof (int));

  printf("Preprocessing sequences...\n");
  asizeel = preprocess_rflcs(xo, yo, x, y, mo, no, &m, &n, asize);
  printf("x = ");
  print_sequence(x, m);
  printf("y = ");
  print_sequence(y, n);
  printf("Length of sequence x: %d\n", m);
  printf("Length of sequence y: %d\n", n);
  printf("Preprocessed alphabet size: %d\n", asizeel);
  printf("\n");

  printf("Calculating initial upper bound...\n");
  upper = lcs_cormen(xo, yo, z, mo, no);
  if (asizeel < upper)
    upper = asizeel;
  printf("Upper bound: %d\n\n", upper);

  printf("Calculating initial lower bound...\n");
  lower = rflcs_approx1(x, y, z, m, n, asize);
  k = rflcs_approx2(x, y, zaux, m, n, 1, asize);
  if (k > lower) {
    lower = k;
    copy_sequence(z, zaux, k);
  }
  k = rflcs_approx2(x, y, zaux, m, n, 2, asize);
  if (k > lower) {
    lower = k;
    copy_sequence(z, zaux, k);
  }
  printf("Lower bound: %d\n\n", lower);

  if (lower == upper) {
    printf("Lower bound equals upper bound - no need to process any further."
	   "\n\n");
    return lower;
  }

  xcount = (int *) malloc_chk(asize * sizeof (int));
  ycount = (int *) malloc_chk(asize * sizeof (int));

  cd = malloc_crossdata(m, n, asize);

  for (i = 0; i < m; i++)
    for (j = 0; j < n; j++)
      cd->revcross[i][j] = -1;

  times(&global_t_start);

  count_seq(xcount, ycount, x, y, m, n, asize);

  construct_model_reponly_ip();
  construct_data_ip(x, y, m, n, asize, cd);

  lp = lpx_read_model(MODPATH, DATPATH, NULL);

  lpparm = (glp_smcp *) malloc_chk(sizeof (glp_smcp));
  glp_init_smcp(lpparm);

  glp_simplex(lp, lpparm);

  ipparm = (glp_iocp *) malloc_chk(sizeof (glp_iocp));
  if (TIME_LIMIT >= 0)
    ipparm->tm_lim = TIME_LIMIT * 1000;
  glp_init_iocp(ipparm);

  info = create_info(x, y, z, m, n, asize, cd, xcount, ycount, lower);
  ipparm->cb_func = callback_bb;
  ipparm->cb_info = info;

  glp_intopt(lp, ipparm);

  lower = convert_lp_to_sequence(lp, x, y, z, cd, 1);

  glp_delete_prob(lp);

  free(x);
  free(y);
  free(zaux);
  free(xcount);
  free(ycount);
  
  free_crossdata(cd, m);

  return lower;
}

int relax_cut_loop (glp_prob *lp, int *x, int *y, int *z, int m, int n, 
                    int asize, crossdata_t *cd, int *xcount, int *ycount,
                    int *heuristic)
{
  int i, j, k, row, lower, old_lower, print, iter;
  int *zi, *zj, *ind;
  double **weight, *val, wsum, time_elapsed;
  glp_smcp *lpparm;

  ind = (int *) malloc_chk((MIN(m, n) + 1) * sizeof (int));
  val = (double *) malloc_chk((MIN(m, n) + 1) * sizeof (double));
  zi = (int *) malloc_chk(m * sizeof (int));
  zj = (int *) malloc_chk(n * sizeof (int));
  weight = (double **) malloc_chk(m * sizeof (double *));
  for (i = 0; i < m; i++)
    weight[i] = (double *) malloc_chk(n * sizeof (double));

  for (i = 0; i < m; i++)
    for (j = 0; j < n; j++)
      weight[i][j] = -1;

  print = 0;

  lpparm = (glp_smcp *) malloc_chk(sizeof (glp_smcp));
  glp_init_smcp(lpparm);

  lower = 0;
  iter = 0;

  lpparm->msg_lev = GLP_MSG_ERR;
  if (TIME_LIMIT >= 0)
    lpparm->tm_lim = TIME_LIMIT * 1000;
  printf("Separation begins...\n");
  /* ensure the next simplex is primal */
  lpparm->meth = GLP_PRIMAL;

  glp_simplex(lp, lpparm);
  iter++;

  lower = prob_approx(lp, x, y, z, m, n, asize, xcount, ycount, cd, lower);
  old_lower = lower;

  for (i = 1; i <= cd->csize; i++)
    weight[cd->crossx[i-1]][cd->crossy[i-1]] = glp_get_col_prim(lp, i);

  k = max_cross_2(x, y, zi, zj, m, n, cd, weight, &wsum);

  print_status(iter, wsum, lower, glp_get_obj_val(lp));

  /* toggles dual simplex mode for the next simplex runs */
  lpparm->meth = GLP_DUALP;

  while (wsum > 1 + EPSILON) {
    /* adds a new restriction acquired by separation */
    glp_add_rows(lp, 1);
    row = glp_get_num_rows(lp);
    for (i = 1; i <= k; i++) {
      ind[i] = cd->revcross[zi[i-1]][zj[i-1]] + 1;
      if (ind[i] == 0) {
        fprintf(stderr, "Error: Index unexpected.\n");
        exit(EXIT_FAILURE);
      }
      val[i] = 1;
    }
    glp_set_mat_row(lp, row, k, ind, val);
    glp_set_row_bnds(lp, row, GLP_UP, 0, 1);

    /* uses dual simplex to reobtain a feasible solution */
    glp_simplex(lp, lpparm);
    iter++;

    lower = prob_approx(lp, x, y, z, m, n, asize, xcount, ycount, cd, lower);

    if (lower > old_lower) {
      old_lower = lower;
      print = 1;
    }

    /* acquires another maximal crossing based on last simplex */
    for (i = 1; i <= cd->csize; i++)
      weight[cd->crossx[i-1]][cd->crossy[i-1]] = glp_get_col_prim(lp, i);

    k = max_cross_2(x, y, zi, zj, m, n, cd, weight, &wsum);

    if (print == 1 || iter % 100 == 0 ||
        (int) (glp_get_obj_val(lp) + EPSILON) == lower) {
      print_status(iter, wsum, lower, glp_get_obj_val(lp));
      print = 0;
    }

    if ((int) (glp_get_obj_val(lp) + EPSILON) == lower) {
      printf("Heuristic is an integer optimal solution for subproblem.\n");
      printf("Separation done.\n");
      *heuristic = lower;
      return 1;
    }

    times(&global_t_end);
    time_elapsed = (((global_t_end.tms_utime + global_t_end.tms_cutime) -
		     (global_t_start.tms_utime + global_t_start.tms_cutime)) /
		    sysconf(_SC_CLK_TCK));
    if (TIME_LIMIT >= 0 && time_elapsed >= TIME_LIMIT) {
      printf("%5d: Time limit reached.\n", iter);
      return -1;
    }
  }

  if (iter > 1)
    print_status(iter, wsum, lower, glp_get_obj_val(lp));
  printf("Separation done.\n");
  *heuristic = lower;

  free(ind);
  free(val);
  free(zi);
  free(zj);
  for (i = 0; i < m; i++)
    free(weight[i]);
  free(weight);

  return 0;
}

/* Finds a maximal crossing using a modified version of the LCS algorithm.
 */
int max_cross (int *x, int *y, int *zi, int *zj, int m, int n, 
               double **weight, double *wsum)
{
  int i, j, k;
  double **c;
  char **b;

  c = (double **) malloc_chk((m + 1) * sizeof (double *));
  for (i = 0; i < m + 1; i++)
    c[i] = (double *) malloc_chk((n + 1) * sizeof (double));

  b = (char **) malloc_chk((m + 1) * sizeof (char *));
  for (i = 0; i < m + 1; i++)
    b[i] = (char *) malloc_chk((n + 1) * sizeof (char));

  for (i = 1; i <= m; i++)
    c[i][0] = 0;
  for (j = 0; j <= n; j++)
    c[0][j] = 0;

  for (i = 1; i <= m; i++) {
    for (j = 1; j <= n; j++) {
      if (x[i-1] == y[n-j]) {
	c[i][j] = c[i-1][j-1] + weight[i-1][n-j];
	b[i][j] = 'd';
        if (c[i][j] < c[i-1][j]) {
          c[i][j] = c[i-1][j];
          b[i][j] = 'u';          
        }
        else if (c[i][j] < c[i][j-1]) {
          c[i][j] = c[i][j-1];
          b[i][j] = 'l';
        }
      }
      else if (c[i-1][j] >= c[i][j-1]) {
	c[i][j] = c[i-1][j];
	b[i][j] = 'u';
      }
      else {
	c[i][j] = c[i][j-1];
	b[i][j] = 'l';
      }
    }
  }

  k = 0;
  i = m;
  j = n;

  while (i != 0 && j != 0) {
    if (b[i][j] == 'd') {
      zi[k] = i - 1;
      zj[k] = n - j;
      i--;
      j--;
      k++;
    }
    else if (b[i][j] == 'u')
      i--;
    else
      j--;
  }

  *wsum = c[m][n];

  for (i = 0; i < m + 1; i++) {
    free(b[i]);
    free(c[i]);
  }
  free(b);
  free(c);

  return k;
}

/* Finds a group of alignments in which for every alignment a and b, a crosses
 * with b or a has the same symbol as b. Uses a modified version of the LCS
 * algorithm.
 */
int max_cross_2 (int *x, int *y, int *zi, int *zj, int m, int n,
		 crossdata_t *cd, double **weight, double *wsum)
{
  int i, j, k, l, s, t, inew, jnew;
  int **b;
  double **c;
  double sum;

  c = (double **) malloc_chk((m + 1) * sizeof (double *));
  for (i = 0; i < m + 1; i++)
    c[i] = (double *) malloc_chk((n + 1) * sizeof (double));

  b = (int **) malloc_chk((m + 1) * sizeof (int *));
  for (i = 0; i < m + 1; i++)
    b[i] = (int *) malloc_chk((n + 1) * sizeof (int));

  for (i = 1; i <= m; i++)
    c[i][0] = 0;
  for (j = 0; j <= n; j++)
    c[0][j] = 0;

  for (i = 1; i <= m; i++) {
    for (j = 1; j <= n; j++) {
      c[i][j] = c[i-1][j];
      b[i][j] = -1; /* up */
      
      if (c[i][j] < c[i][j-1]) {
	c[i][j] = c[i][j-1];
	b[i][j] = -2; /* left */
      }

      if (x[i-1] == y[n-j]) {
	s = x[i-1];
	for (k = cd->scross[s]; k < cd->csize && x[cd->crossx[k]] == s; k++) {
	  if (cd->crossx[k] <= i - 1 && cd->crossy[k] >= n - j) {
	    sum = c[cd->crossx[k]][n - cd->crossy[k] - 1];

	    /* sum of the weights between the two alignments */
	    for (l = cd->scross[s]; l < cd->csize && x[cd->crossx[l]] == s; 
		 l++) {
	      if (cd->crossx[l] <= i - 1 && cd->crossy[l] >= n - j &&
		  cd->crossx[l] >= cd->crossx[k] && 
		  cd->crossy[l] <= cd->crossy[k])
		sum += weight[cd->crossx[l]][cd->crossy[l]];
	    }

	    if (sum > c[i][j]) {
	      c[i][j] = sum;
	      b[i][j] = k;
	    }
	  }
	}
      }
    }
  }

  k = 0;
  i = m;
  j = n;

  while (i != 0 && j != 0) {
    if (b[i][j] == -1) /* up */
      i--;
    else if (b[i][j] == -2) /* left */
      j--;
    else { /* b[i][j] holds the index of the alignment */
      l = b[i][j];
      inew = cd->crossx[l];
      jnew = n - cd->crossy[l] - 1;
      s = x[i-1];
      for (t = cd->scross[s]; t < cd->csize && x[cd->crossx[t]] == s; t++) {
	if (cd->crossx[t] <= i - 1 && cd->crossy[t] >= n - j &&
	    cd->crossx[t] > inew - 1 && cd->crossy[t] < n - jnew) {
	  zi[k] = cd->crossx[t];
	  zj[k] = cd->crossy[t];
	  k++;
	}
      }
      i = inew;
      j = jnew;
    }
  }

  *wsum = c[m][n];

  for (i = 0; i < m + 1; i++) {
    free(b[i]);
    free(c[i]);
  }
  free(b);
  free(c);

  return k;
}

int max_cross_3 (int *x, int *y, int *zi, int *zj, int m, int n,
		 crossdata_t *cd, double **weight, double *wsum,
		 glp_prob *lp)
{
  int i, j, k, l, s, t, inew, jnew, ii, jj, row;
  int *done, *ind;
  int **b;
  double sum;
  double *val;
  double **c;

  done = (int *) malloc_chk(cd->csize * sizeof (int));
  ind = (int *) malloc_chk((MIN(m, n) + 1) * sizeof (int));
  val = (double *) malloc_chk((MIN(m, n) + 1) * sizeof (double));

  c = (double **) malloc_chk((m + 1) * sizeof (double *));
  for (i = 0; i < m + 1; i++)
    c[i] = (double *) malloc_chk((n + 1) * sizeof (double));

  b = (int **) malloc_chk((m + 1) * sizeof (int *));
  for (i = 0; i < m + 1; i++)
    b[i] = (int *) malloc_chk((n + 1) * sizeof (int));

  for (i = 0; i < cd->csize; i++)
    done[i] = 0;

  for (i = 1; i <= m; i++)
    c[i][0] = 0;
  for (j = 0; j <= n; j++)
    c[0][j] = 0;

  for (i = 1; i <= m; i++) {
    for (j = 1; j <= n; j++) {
      c[i][j] = c[i-1][j];
      b[i][j] = -1; /* up */
      
      if (c[i][j] < c[i][j-1]) {
	c[i][j] = c[i][j-1];
	b[i][j] = -2; /* left */
      }

      if (x[i-1] == y[n-j]) {
	s = x[i-1];
	for (k = cd->scross[s]; k < cd->csize && x[cd->crossx[k]] == s; k++) {
	  if (cd->crossx[k] <= i - 1 && cd->crossy[k] >= n - j) {
	    sum = c[cd->crossx[k]][n - cd->crossy[k] - 1];

	    /* sum of the weights between the two alignments */
	    for (l = cd->scross[s]; l < cd->csize && x[cd->crossx[l]] == s; 
		 l++) {
	      if (cd->crossx[l] <= i - 1 && cd->crossy[l] >= n - j &&
		  cd->crossx[l] >= cd->crossx[k] && 
		  cd->crossy[l] <= cd->crossy[k])
		sum += weight[cd->crossx[l]][cd->crossy[l]];
	    }

	    if (sum > c[i][j]) {
	      c[i][j] = sum;
	      b[i][j] = k;
	    }
	  }
	}
      }
    }
  }

  for (ii = m; ii > 0; ii--)
    for (jj = n; jj > 0; jj--) {
      if (c[ii][jj] < 1 + EPSILON)
	break;
      if (b[ii][jj] >= 0 && done[b[ii][jj]] == 0) {
	i = ii;
	j = jj;
	k = 0;

	while (i != 0 && j != 0) {
	  if (b[i][j] == -1) /* up */
	    i--;
	  else if (b[i][j] == -2) /* left */
	    j--;
	  else { /* b[i][j] holds the index of the alignment */
	    l = b[i][j];
	    inew = cd->crossx[l];
	    jnew = n - cd->crossy[l] - 1;
	    s = x[i-1];
	    for (t = cd->scross[s]; t < cd->csize && 
		   x[cd->crossx[t]] == s; t++) {
	      if (cd->crossx[t] <= i - 1 && cd->crossy[t] >= n - j &&
		  cd->crossx[t] > inew - 1 && cd->crossy[t] < n - jnew) {
		zi[k] = cd->crossx[t];
		zj[k] = cd->crossy[t];
		k++;
	      }
	    }
	    i = inew;
	    j = jnew;
	    done[l] = 1;
	  }
	}

	glp_add_rows(lp, 1);
	row = glp_get_num_rows(lp);
	for (i = 1; i <= k; i++) {
	  ind[i] = cd->revcross[zi[i-1]][zj[i-1]] + 1;
	  if (ind[i] == 0) {
	    fprintf(stderr, "Error: Index unexpected.\n");
	    exit(EXIT_FAILURE);
	  }
	  val[i] = 1;
	}
	glp_set_mat_row(lp, row, k, ind, val);
	glp_set_row_bnds(lp, row, GLP_UP, 0, 1);
      }
    }

  *wsum = c[m][n];

  for (i = 0; i < m + 1; i++) {
    free(b[i]);
    free(c[i]);
  }
  free(b);
  free(c);
  free(done);

  return k;
}

int prob_approx (glp_prob *lp, int *x, int *y, int *z, int m, int n,
		 int asize, int *xcount, int *ycount, crossdata_t *cd,
		 int kmax)
{
  int i, j, k, maux, naux;
  int *xchoice, *ychoice;
  int *xaux, *yaux, *zaux;
  double **prob;

  xchoice = (int *) malloc_chk(asize * sizeof (int));
  ychoice = (int *) malloc_chk(asize * sizeof (int));
  xaux = (int *) malloc_chk(m * sizeof (int));
  yaux = (int *) malloc_chk(n * sizeof (int));
  zaux = (int *) malloc_chk(MIN(m, n) * sizeof (int));
  prob = (double **) malloc_chk(asize * sizeof (double *));
  for (i = 0; i < asize; i++)
    prob[i] = (double *) malloc_chk(MAX(m, n) * sizeof (double));

  for (i = 0; i < PROB_TIMES; i++) {
    construct_prob_lp(lp, x, y, m, n, xcount, ycount, asize, cd, prob);

    for (j = 0; j < asize; j++)
      xchoice[j] = ychoice[j] = 0;

    choice_3(xcount, ycount, xchoice, ychoice, asize, prob);

    maux = choose_seq(x, xaux, xchoice, m);
    naux = choose_seq(y, yaux, ychoice, n);

    k = lcs_cormen(xaux, yaux, zaux, maux, naux);

    if (k > kmax) {
      kmax = k;
      for (j = 0; j < k; j++)
        z[j] = zaux[j];
    }
  }

  free(xchoice);
  free(ychoice);
  free(xaux);
  free(yaux);
  free(zaux);

  for (i = 0; i < asize; i++)
    free(prob[i]);
  free(prob);

  return kmax;
}

void callback_bb (glp_tree *tree, void *info)
{
  int i, k, row, h;
  int ind[2];
  double val[2];
  double *vars;
  info_t *inf;
  glp_prob *lp;

  inf = (info_t *) info;

  if (inf->init_lower > 0) {
    printf("Adding initial heuristic...\n");
    vars = (double *) malloc_chk((inf->cd->csize + 1) * sizeof (double));
    convert_sequence_to_vars(inf->x, inf->y, inf->z, inf->init_lower, vars, 
                             inf->cd);
    glp_ios_heur_sol(tree, vars);
    inf->init_lower = 0;
  }

  switch (glp_ios_reason(tree)) {
    case GLP_IROWGEN:
      h = 0;
      lp = glp_ios_get_prob(tree);
      k = relax_cut_loop(lp, inf->x, inf->y, inf->z, inf->m, inf->n,
                         inf->asize, inf->cd, inf->xcount, inf->ycount, &h);

      vars = (double *) malloc_chk((inf->cd->csize + 1) * sizeof (double));
      if (k == 1) {
        /* restricts the subproblem to its integer optimal solution found
         * by the heuristic */
        convert_sequence_to_vars(inf->x, inf->y, inf->z, h, vars, inf->cd);
        val[1] = 1;
        glp_add_rows(lp, inf->cd->csize);
        for (i = 1; i <= inf->cd->csize; i++) {
          row = glp_get_num_rows(lp) - (i - 1);
          ind[1] = i;
          glp_set_mat_row(lp, row, 1, ind, val);
          glp_set_row_bnds(lp, row, GLP_FX, vars[i], vars[i]);
        }
      }
      else if (k == -1) {
	glp_ios_terminate(tree);
      }
      else {
        convert_sequence_to_vars(inf->x, inf->y, inf->z, h, vars, inf->cd);
        glp_ios_heur_sol(tree, vars);
      }
      break;
    /*
    case GLP_ICUTGEN:
      if (generate_cuts(lp, LPX_C_ALL) != LPX_E_OK) {
	fprintf(stderr, "Error: Unable to generate cutting planes.\n");
	exit(EXIT_FAILURE);
      }
      break;
    */
  }
  return;
}

info_t *create_info (int *x, int *y, int *z, int m, int n, int asize,
		     crossdata_t *cd, int *xcount, int *ycount, int lower)
{
  info_t *info;

  info = (info_t *) malloc_chk(sizeof (info_t));
  info->x = x;
  info->y = y;
  info->z = z;
  info->m = m;
  info->n = n;
  info->asize = asize;
  info->cd = cd;
  info->xcount = xcount;
  info->ycount = ycount;
  info->init_lower = lower;

  return info;
}

void print_status (int iter, double wsum, int lower, double obj_val)
{
  printf("%5d: ", iter);
  printf("wsum = %.4f\t", wsum);
  printf("lb (heuristic) = %3d\t", lower);
  printf("ub (simplex) = %.5f\n", obj_val);
  fflush(NULL);
}
