/* util.c
 * Utility functions
 */

#include "util.h"

/* Given an array a and an element n, does a binary search on a for n except 
 * that, if the element was not found, returns the index of the element 
 * immediately higher.
 */
int bsearchh (double n, double *a, int asize)
{
  int l, r, k;
  l = 0;
  r = asize - 1;
  while (l <= r) {
    k = l + (r - l) / 2;
    if (n < a[k])
      r = k - 1;
    else if (n > a[k])
      l = k + 1;
    else
      return k;
  }
  return l;
}

/* Allocates memory and checks for errors.
 */
void *malloc_chk (unsigned int n)
{
  void *ptr;
  if ((ptr = malloc(n)) == NULL) {
    fprintf(stderr, "Error: Unable to allocate memory.\n");
    exit(EXIT_FAILURE);
  }
  return ptr;
}

/* Converts a string to integer and checks for errors.
 */
int atoi_chk (const char *nptr)
{
  char *endptr;
  int n;

  n = strtol(nptr, &endptr, 10);
  if (endptr == nptr) {
    fprintf(stderr, "Error: Unable to parse number.\n");
    exit(EXIT_FAILURE);
  }

  return n;
}

/* Prints the sequence x of size n.
 */
void print_sequence (int *x, int n)
{
  int i;

  for (i = 0; i < n; i++)
    printf("%d ", x[i]);
  printf("\n");
}

/* Copies src of size n to dest.
 */
void copy_sequence (int *dest, int *src, int n)
{
  memcpy(dest, src, n * sizeof (int));
}

/* Prints the usage of the program.
 */
void print_usage (char *binpath)
{
  printf("Usage: %s [options] XLENGTH YLENGTH\n\n", binpath);
  printf("Options:\n");
  printf("   -a value, --alphabet value\n"
         "\tdefine the alphabet size to value [default %d]\n", DEFAULT_ASIZE);
  printf("   -m filename, --metapost filename\n"
         "\tstore a metapost file representing the result on filename \n"
         "\t(by default no file is written)\n");
  printf("   -g number, --generate number\n"
         "\tchoose the method which the sequences are generated: \n"
         "\t   1 - uniformly [default]\n"
         "\t   2 - half of the alphabet has higher probability\n"
	 "\t   3 - XLENGTH and YLENGTH are number of repetitions per symbol\n"
	 "\t       instead of lengths\n");
  printf("   -c type, --color type\n"
         "\tcolorize result on metapost:\n"
	 "\t   0 - no color\n"
	 "\t   1 - greyscale\n"
	 "\t   2 - hue variation [default]\n");
  printf("   -s, --silent\n"
         "\tsilent mode, outputs only lengths (note: silent is more of a\n"
         "\thack - it redirects output to /dev/null, thus is not portable)\n");
  printf("   --seed\n"
         "\tdefine the seed for the pseudorandom number generator\n");
  printf("   -i, --integer\n"
         "\trun integer programming algorithm using only glpk formulation\n");
  printf("   --derand3\n"
	 "\trun a derandomized version of approximation algorithm 3\n");
  printf("   --mgraph\n"
         "\tonly works in conjunction with -m, --metapost, stores the graph\n"
         "\tin the metapost file\n");
  printf("   --approx4\n"
         "\tapprox. algorithm using simplex to decide seq. to apply lcs\n");
  printf("   -b, --buildgraph\n"
         "\tbuilds the cross/noncross graph (works in conjunction with\n"
         "\tother options that require this graph)\n");
  printf("   -l, --load\n"
         "\tloads a file containing a sequence, XLENGTH and YLENGTH are\n"
         "\tignored (the file must contain the alphabet size, XLENGTH,\n"
         "\tYLENGTH, X and Y, in this order)\n");
}

/* Computes the greatest common divisor of x and y.
 */
unsigned int gcd_2 (unsigned int x, unsigned int y)
{
  unsigned int tmp;

  while (y != 0) {
    tmp = y;
    y = x % y;
    x = tmp;
  }

  return x;
}

/* Computes the least common multiple of x and y.
 */
unsigned int lcm_2 (unsigned int x, unsigned int y)
{
  unsigned int tmp;
  tmp = y / gcd_2(x, y);
  check_overflow(x, tmp);
  return (x * tmp);
}

/* Computes the least common multiple of the elements of an array, 
 * ignoring zeros.
 */
unsigned int lcm_array (int *a, int asize)
{
  int i;
  unsigned int lcm;

  lcm = 1;
  for (i = 0; i < asize; i++)
    if (a[i] > 1)
      lcm = lcm_2(a[i], lcm);

  return lcm;
}

/* Compare function for qsort.
 */
int int_compare (const void *p1, const void *p2)
{
  int *x, *y;
  x = (int *) p1;
  y = (int *) p2;
  return (*x - *y);
}

/* Returns error if the multiplication of x and y results in an overflow.
 */
void check_overflow (unsigned int x, unsigned int y)
{
  if (y != 0 && x > (UINT_MAX / y)) {
    fprintf(stderr, "Error: Integer overflow.\n");
    exit(EXIT_FAILURE);
  }
}
